bzr_plugin_name = 'fastimport'

bzr_commands = [
    "fast-import",
    "fast-import-filter",
    "fast-import-info",
    "fast-import-query",
    "fast-export",
    ]

bzr_plugin_version = (0, 14, 0, 'dev', 0)
bzr_minimum_version = (2, 0, 0)
bzr_maximum_version = None
